from django.conf import settings
from django.utils import timezone
from django.utils.encoding import force_text
from django.utils.translation import ugettext_lazy as _l

from chamber.utils.datastructures import ChoicesEnum

from pymess.backend.sms import SMSBackend
from pymess.models import OutputSMSMessage

from twilio.rest import TwilioRestClient


class TwilioSMSBackend(SMSBackend):
    """
    SMS backend implementing twilio service https://www.twilio.com/
    """

    twilio_client = None

    STATE = ChoicesEnum(
        ('ACCEPTED', _l('accepted'), 'accepted'),
        ('QUEUED', _l('queued'), 'queued'),
        ('SENDING', _l('sending'), 'sending'),
        ('SENT', _l('sent'), 'sent'),
        ('DELIVERED', _l('delivered'), 'delivered'),  # TODO implement checking delivery status
        ('RECEIVED', _l('received'), 'received'),
        ('FAILED', _l('failed'), 'failed'),
        ('UNDELIVERED', _l('undelivered'), 'undelivered'),
    )

    STATES_MAPPING = {
        STATE.ACCEPTED: OutputSMSMessage.STATE.SENT,
        STATE.QUEUED: OutputSMSMessage.STATE.SENT,
        STATE.SENDING: OutputSMSMessage.STATE.SENT,
        STATE.SENT: OutputSMSMessage.STATE.SENT,
        STATE.DELIVERED: OutputSMSMessage.STATE.DELIVERED,
        STATE.RECEIVED: OutputSMSMessage.STATE.DELIVERED,
        STATE.FAILED: OutputSMSMessage.STATE.ERROR,
        STATE.UNDELIVERED: OutputSMSMessage.STATE.ERROR,
    }

    def _get_twilio_client(self):
        """
        Connect to the twilio service
        """
        if not self.twilio_client:
            self.twilio_client = TwilioRestClient(settings.TWILIO_ACCOUNT_SID, settings.TWILIO_AUTH_TOKEN)
        return self.twilio_client

    def publish_message(self, message):
        """
        Method uses twilio REST client for sending SMS message
        :param message: SMS message
        """
        client = self._get_twilio_client()
        try:
            result = client.messages.create(
                from_=settings.TWILIO_SENDER,
                to=force_text(message.recipient),
                body=message.content
            )
            self.update_message(
                message,
                state=self.STATES_MAPPING[result.status],
                error=result.error_message if result.error_message else None,
                sent_at=timezone.now()
            )
        except Exception as ex:
            self.update_message(message, state=OutputSMSMessage.STATE.ERROR, error=force_text(ex))



# -*- coding: utf-8 -*-
# Generated by Django 1.10.1 on 2018-01-26 10:07
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='UserDevice',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at')),
                ('registration_id', models.CharField(db_index=True, max_length=256, verbose_name='registration ID')),
                ('is_active', models.BooleanField(default=True, verbose_name='is active')),
                ('device_id', models.CharField(max_length=255, verbose_name='device ID')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE,
                                           related_name='user_devices', to=settings.AUTH_USER_MODEL,
                                           verbose_name='user')),
            ],
            options={
                'ordering': ('-created_at',),
                'verbose_name_plural': 'user devices',
                'verbose_name': 'user device',
            },
        ),
    ]



from pymess.backend.emails import EmailBackend
from pymess.models import EmailMessage


class DummyEmailBackend(EmailBackend):
    """
    Dummy e-mail backend used for testing environments. Backend only logs messages to the database.
    """

    def publish_message(self, message):
        self.update_message(message, state=EmailMessage.STATE.DEBUG)



from django.core.management.base import BaseCommand

from pymess.config import get_email_sender


class Command(BaseCommand):
    """
    Command for sending e-mails in batch.
    """

    def handle(self, *args, **kwargs):
        get_email_sender().send_batch()



from django.db.models import Manager
from django.contrib.contenttypes.models import ContentType

from pymess.utils import has_int_pk


class RelatedObjectManager(Manager):

    def create_from_related_object(self, related_object):
        return self.create(
            object_id_int=related_object.pk if has_int_pk(related_object) else None,
            object_id=related_object.pk,
            content_type=ContentType.objects.get_for_model(related_object)
        )

    def create_from_related_objects(self, *related_objects):
        return [
            self.create_from_related_object(related_object)
            for related_object in related_objects
        ]



from jsonfield.fields import JSONField

import six

from django.core.exceptions import ValidationError
from django.contrib.contenttypes.models import ContentType
from django.contrib.contenttypes.fields import GenericForeignKey
from django.db import models
from django.utils.translation import ugettext, ugettext_lazy as _
from django.utils.encoding import force_text
from django.template import Template, Context
from django.template.exceptions import TemplateSyntaxError, TemplateDoesNotExist

from chamber.models import SmartModel
from chamber.utils import remove_accent
from chamber.utils.datastructures import ChoicesNumEnum

from pymess.config import settings, get_sms_sender
from pymess.utils import normalize_phone_number

from .common import RelatedObjectManager


__ALL__ = (
    'OutputSMSMessage',
    'OutputSMSRelatedObjects',
    'AbstractSMSTemplate',
    'AbstractEmailTemplate',
    'SMSTemplate',
)


class OutputSMSMessage(SmartModel):

    STATE = ChoicesNumEnum(
        ('WAITING', _('waiting'), 1),
        ('UNKNOWN', _('unknown'), 2),
        ('SENDING', _('sending'), 3),
        ('SENT', _('sent'), 4),
        ('ERROR', _('error'), 5),
        ('DEBUG', _('debug'), 6),
        ('DELIVERED', _('delivered'), 7),
    )

    sent_at = models.DateTimeField(verbose_name=_('sent at'), null=True, blank=True, editable=False)
    sender = models.CharField(verbose_name=_('sender'), null=True, blank=True, max_length=20)
    recipient = models.CharField(verbose_name=_('recipient'), null=False, blank=False, max_length=20)
    content = models.TextField(verbose_name=_('content'), null=False, blank=False, max_length=700)
    template_slug = models.SlugField(verbose_name=_('slug'), max_length=100, null=True, blank=True, editable=False)
    template = models.ForeignKey(settings.SMS_TEMPLATE_MODEL, verbose_name=_('template'), blank=True, null=True,
                                 on_delete=models.SET_NULL, related_name='output_sms_messages')
    state = models.IntegerField(verbose_name=_('state'), null=False, blank=False, choices=STATE.choices, editable=False)
    backend = models.CharField(verbose_name=_('backend'), null=True, blank=True, editable=False, max_length=250)
    error = models.TextField(verbose_name=_('error'), null=True, blank=True, editable=False)
    extra_data = JSONField(verbose_name=_('extra data'), null=True, blank=True, editable=False)
    extra_sender_data = JSONField(verbose_name=_('extra sender data'), null=True, blank=True, editable=False)
    tag = models.SlugField(verbose_name=_('tag'), null=True, blank=True, editable=False)

    def clean_recipient(self):
        self.recipient = normalize_phone_number(force_text(self.recipient))

    def clean_content(self):
        if not settings.SMS_USE_ACCENT:
            self.content = six.text_type(remove_accent(six.text_type(self.content)))

    @property
    def failed(self):
        return self.state == self.STATE.ERROR

    def __str__(self):
        return self.recipient

    class Meta:
        verbose_name = _('output SMS')
        verbose_name_plural = _('output SMS')
        ordering = ('-created_at',)


class OutputSMSRelatedObject(SmartModel):

    output_sms_message = models.ForeignKey(OutputSMSMessage, verbose_name=_('output SMS message'), null=False,
                                           blank=False, on_delete=models.CASCADE, related_name='related_objects')
    content_type = models.ForeignKey(ContentType, verbose_name=_('content type of the related object'),
                                     null=False, blank=False, on_delete=models.CASCADE)
    object_id = models.TextField(verbose_name=_('ID of the related object'), null=False, blank=False)
    object_id_int = models.PositiveIntegerField(verbose_name=_('ID of the related object in int format'), null=True,
                                                blank=True, db_index=True)
    content_object = GenericForeignKey('content_type', 'object_id')

    objects = RelatedObjectManager()

    class Meta:
        verbose_name = _('related object of a SMS message')
        verbose_name_plural = _('related objects of SMS messages')
        ordering = ('-created_at',)


class AbstractSMSTemplate(SmartModel):

    slug = models.SlugField(verbose_name=_('slug'), max_length=100, null=False, blank=False, editable=False,
                            primary_key=True)
    body = models.TextField(verbose_name=_('message body'), null=True, blank=False)

    def clean_body(self, context_data=None):
        try:
            self.render_body(context_data or {})
        except (TemplateSyntaxError, TemplateDoesNotExist) as ex:
            raise ValidationError(ugettext('Error during template body rendering: "{}"').format(ex))

    def get_body(self):
        return self.body

    def _update_context_data(self, context_data):
        return context_data

    def render_body(self, context_data):
        context_data = self._update_context_data(context_data)
        return Template(self.get_body()).render(Context(context_data))

    def can_send(self, recipient, context_data):
        return True

    def send(self, recipient, context_data, related_objects=None, tag=None):
        if self.can_send(recipient,context_data):
            return get_sms_sender().send(
                recipient,
                self.render_body(context_data),
                template=self,
                related_objects=related_objects,
                tag=tag
            )
        else:
            return None

    def __str__(self):
        return self.slug

    class Meta:
        abstract = True
        verbose_name = _('SMS template')
        verbose_name_plural = _('SMS templates')
        ordering = ('-created_at',)


class SMSTemplate(AbstractSMSTemplate):
    pass



"""
If the django-security library is present in the system, we want to log HTTP requests
using its modified get(), post() etc. functions instead of standard the requests library.
"""

try:
    from security.transport.security_requests import *
except ImportError:
    from requests import *
    from requests import (get as _get, options as _options, head as _head, post as _post, put as _put, patch as _patch,
                          delete as _delete)

    def get(url, slug=None, related_objects=None, **kwargs):
        return _get(url, **kwargs)

    def options(url, slug=None, related_objects=None, **kwargs):
        return _options(url, **kwargs)

    def head(url, slug=None, related_objects=None, **kwargs):
        return _head(url, **kwargs)

    def post(url, slug=None, related_objects=None, **kwargs):
        return _post(url, **kwargs)

    def put(url, slug=None, related_objects=None, **kwargs):
        return _put(url, **kwargs)

    def patch(url, slug=None, related_objects=None, **kwargs):
        return _patch(url, **kwargs)

    def delete(url, slug=None, related_objects=None, **kwargs):
        return _delete(url, **kwargs)



import codecs
import os

from django.core.exceptions import ObjectDoesNotExist
from django.core.management.base import BaseCommand

from pymess.config import get_email_template_model, settings


class Command(BaseCommand):
    """
    Command synchronize e-mails body from directory.
    Every e-mail body is named with e-mail slug and stored as a HTML file.
    """

    def handle(self, *args, **options):
        verbosity = int(options.get('verbosity'))
        if verbosity > 0:
            self.stdout.write('Syncing e-mails')

        files_with_email_html_body = [
            filename for filename in os.listdir(
                os.path.join(settings.EMAIL_HTML_DATA_DIRECTORY)
            ) if filename.endswith('.html')
        ]
        for filename in files_with_email_html_body:
            try:
                template_obj = get_email_template_model().objects.get(slug=filename[:-5])
                with codecs.open(os.path.join(directory, filename), 'r', encoding='utf-8-sig') as file:
                    template_obj.change_and_save(body=file.read())
            except ObjectDoesNotExist:
                if verbosity > 0:
                    self.stderr.write('Template model with slug "{}" does not exists'.format(filename[:-5]))
        if verbosity > 0:
            self.stdout.write(
                self.style.SUCCESS('Synced "{}" e-mail templates'.format(len(files_with_email_html_body)))
            )



# Generated by Django 2.0 on 2018-06-21 22:31

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('pymess', '0002_auto_20180221_1439'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='emailtemplate',
            name='id',
        ),
        migrations.RemoveField(
            model_name='smstemplate',
            name='id',
        ),
        migrations.AlterField(
            model_name='emailtemplate',
            name='slug',
            field=models.SlugField(editable=False, max_length=100, primary_key=True, serialize=False, verbose_name='slug'),
        ),
        migrations.AlterField(
            model_name='smstemplate',
            name='slug',
            field=models.SlugField(editable=False, max_length=100, primary_key=True, serialize=False, verbose_name='slug'),
        ),
    ]



import logging

from django.utils.encoding import force_text

from chamber.exceptions import PersistenceException

from pymess.config import get_email_template_model, get_email_sender, settings
from pymess.models import EmailMessage
from pymess.utils import fullname
from pymess.lockfile import FileLock, AlreadyLocked, LockTimeout


LOGGER = logging.getLogger(__name__)


class EmailBackend(object):
    """
    Base class for E-mail backend containing implementation of e-mail service that
    is used for sending messages.
    """

    class EmailSendingError(Exception):
        pass

    def _get_extra_sender_data(self):
        """
        Gets arguments that will be saved with the message in the extra_sender_data field.
        """
        return {}

    def _get_extra_message_kwargs(self):
        """
        Gets model message kwargs that will be saved with the message.
        """
        return {}

    def update_message(self, message, extra_sender_data=None, **kwargs):
        """
        Method for updating state of the message after sending
        :param message: e-mail message object
        :param extra_sender_data: extra data that will be saved to the extra_sender_data field
        :param kwargs: changed object kwargs
        :return:
        """
        extra_sender_data = {
            **self._get_extra_sender_data(),
            **({} if extra_sender_data is None else extra_sender_data)
        }
        message.change_and_save(
            backend=fullname(self),
            extra_sender_data=extra_sender_data,
            **kwargs
        )

    def create_message(self, sender, sender_name, recipient, subject, content, related_objects, tag, template,
                       attachments, **email_kwargs):
        """
        Create e-mail which will be logged in the database.
        :param sender: e-mail address of the sender
        :param sender_name: friendly name of the sender
        :param recipient: e-mail address of the receiver
        :param subject: subject of the e-mail message
        :param content: content of the e-mail message
        :param related_objects: list of related objects that will be linked with the e-mail message using generic
        relation
        :param tag: string mark that will be saved with the message
        :param template: template object from which content, subject and sender of the message was created
        :param attachments: list of files that will be sent with the message as attachments
        :param email_kwargs: extra data that will be saved in JSON format in the extra_data model field
        """
        try:
            message = EmailMessage.objects.create(
                sender=sender,
                sender_name=sender_name,
                recipient=recipient,
                content=content,
                subject=subject,
                state=self.get_initial_email_state(recipient),
                extra_data=email_kwargs,
                tag=tag,
                template=template,
                template_slug=template.slug if template else None,
                **self._get_extra_message_kwargs()
            )
            if related_objects:
                message.related_objects.create_from_related_objects(*related_objects)
            if attachments:
                message.attachments.create_from_tripples(*attachments)
            return message
        except PersistenceException as ex:
            raise self.EmailSendingError(force_text(ex))

    def publish_message(self, message):
        """
        Place for implementation logic of sending e-mail message.
        :param message: SMS message instance
        """
        raise NotImplementedError

    def send(self, sender, recipient, subject, content, sender_name=None, related_objects=None, tag=None,
             template=None, attachments=None, **email_kwargs):
        """
        Send e-mail with defined values
        :param sender: e-mail address of the sender
        :param recipient: e-mail address of the receiver
        :param subject: subject of the e-mail message
        :param content: content of the e-mail message
        :param sender_name: friendly name of the sender
        :param related_objects: list of related objects that will be linked with the e-mail message with generic
            relation
        :param tag: string mark that will be saved with the message
        :param template: template object from which content, subject and sender was created
        :param attachments: list of files that will be sent with the message as attachments
        :param email_kwargs: extra data that will be saved in JSON format in the extra_data model field
        """
        message = self.create_message(
            sender, sender_name, recipient, subject, content, related_objects, tag, template, attachments, **email_kwargs
        )
        if not settings.EMAIL_BATCH_SENDING:
            self.publish_message(message)
        return message

    def get_initial_email_state(self, recipient):
        """
        returns initial state for logged e-mail message.
        :param recipient: e-mail address of the recipient
        """
        return EmailMessage.STATE.WAITING

    def send_batch(self):
        """
        Method for sending e-mails in a batch mode.
        """
        if not settings.EMAIL_BATCH_SENDING:
            raise self.EmailSendingError('Batch sending is turned off')

        lock = FileLock(settings.EMAIL_BATCH_LOCK_FILE)
        try:
            lock.acquire(settings.EMAIL_BATCH_LOCK_WAIT_TIMEOUT)
        except AlreadyLocked:
            logging.debug("lock already in place. quitting.")
            return
        except LockTimeout:
            logging.debug("waiting for the lock timed out. quitting.")
            return

        sent = 0

        try:
            waiting_emails_qs = EmailMessage.objects.filter(
                state=EmailMessage.STATE.WAITING
            ).order_by('created_at')[:settings.EMAIL_BATCH_SIZE]
            for message in waiting_emails_qs:
                self.publish_message(message)
                sent += 1
        finally:
            logging.debug("releasing lock...")
            lock.release()
            logging.debug("released.")

        logging.info("{} e-mail sent".format(sent))


def send_template(recipient, slug, context_data, related_objects=None, attachments=None, tag=None):
    """
    Helper for building and sending e-mail message from a template.
    :param recipient: e-mail address of the receiver
    :param slug: slug of the e-mail template
    :param context_data: dict of data that will be sent to the template renderer
    :param related_objects: list of related objects that will be linked with the e-mail message with generic
        relation
    :param attachments: list of files that will be sent with the message as attachments
    :param tag: string mark that will be saved with the message 
    :return: e-mail message object or None if template cannot be sent
    """
    return get_email_template_model().objects.get(slug=slug).send(
        recipient,
        context_data,
        related_objects=related_objects,
        attachments=attachments,
        tag=tag
    )


def send(sender, recipient, subject, content, sender_name=None, related_objects=None, attachments=None, tag=None,
         **email_kwargs):
    """
    Helper for sending e-mail message.
    :param sender: e-mail address of the sender
    :param recipient: e-mail address of the receiver
    :param subject: subject of the e-mail message
    :param content: content of the e-mail message
    :param sender_name: friendly name of the sender
    :param related_objects: list of related objects that will be linked with the e-mail message with generic
        relation
    :param tag: string mark that will be saved with the message
    :param attachments: list of files that will be sent with the message as attachments
    :param email_kwargs: extra data that will be saved in JSON format in the extra_data model field
    :return: True if e-mail was successfully sent or False if e-mail is in error state
    """
    return get_email_sender().send(
        sender,
        recipient,
        subject,
        content,
        sender_name=sender_name,
        related_objects=related_objects,
        tag=tag,
        attachments=attachments,
        **email_kwargs
    ).failed


from __future__ import unicode_literals

from django.utils.encoding import force_text

from germanium.rest import RESTTestCase
from germanium.tools import assert_equal

from sender.models import InputSMS


class InputSMSTestCase(RESTTestCase):

    API_URL = '/api/atsinputsmsmessage/'
    ATS_SMS_POST_PAYLOAD = '<?xml version="1.0" encoding="UTF-8" ?><messages>{}</messages>'

    VALID_MESSAGES = """
        <sms uniq="2" sender="+420731545945" recipient="9001103" okey="T2" opid="tmsms" opmid=""
            ts="2006-04-10 09:32:18">test2</sms>
        <sms uniq="3" sender="+420731545945" recipient="9001103" okey="T2" opid="tmsms" opmid=""
            ts="2006-04-10 09:32:23">test3</sms>
        <sms uniq="4" sender="+420731545945" recipient="9001103" okey="T2" opid="tmsms" opmid=""
            ts="2006-04-10 09:32:23"></sms>
    """
    INVALID_MESSAGES = """
        <sms uniq="" sender="+420731545945" recipient="9001103" okey="T2" opid="tmsms" opmid=""
            ts="2006-04-10 09:32:18">test2</sms>
        <sms uniq="invalid" sender="+420731545945" recipient="9001103" okey="T2" opid="tmsms" opmid=""
            ts="2006-04-10 09:32:23">test3</sms>
        <sms uniq="4" sender="+420731545945" recipient="" okey="T2" opid="tmsms" opmid=""
            ts="2006-04-10 09:32:23"></sms>
        <sms uniq="5" sender="+420731545945" recipient="9001103" okey="T2" opid="tmsms" opmid=""
            ts="2006-04-10 09:32:23"></sms>
    """

    def test_ats_request_should_create_input_sms(self):
        sms_count = InputSMS.objects.count()
        response = self.post(self.API_URL, self.ATS_SMS_POST_PAYLOAD.format(self.VALID_MESSAGES))
        self.assert_http_ok(response)  # ATS requires to return 200 in every situation
        assert_equal(sms_count + 3, InputSMS.objects.count())
        assert_equal('test3', InputSMS.objects.get(uniq=3).content)
        assert_equal('', InputSMS.objects.get(uniq=4).content)

    def test_ats_invalid_request_should_return_24_code(self):
        sms_count = InputSMS.objects.count()
        response = self.post(self.API_URL, self.ATS_SMS_POST_PAYLOAD.format(self.INVALID_MESSAGES))
        self.assert_http_ok(response)  # ATS requires to return 200 in every situation
        assert_equal(sms_count + 1, InputSMS.objects.count())
        assert_equal(force_text(response.content).replace('\n', ' ').replace('\r', ''),
                     '<?xml version="1.0" encoding="UTF-8" ?> <status> <code uniq="">24</code> '
                     '<code uniq="invalid">24</code> <code uniq="4">24</code> '
                     '<code uniq="5">23</code> </status>')



# -*- coding: utf-8 -*-
# Generated by Django 1.10.1 on 2018-02-21 13:39
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import jsonfield.fields

from pymess.config import settings


class Migration(migrations.Migration):
    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('pymess', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Attachment',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at')),
                ('file', models.FileField(upload_to='pymess/emails', verbose_name='file')),
                ('content_type', models.CharField(verbose_name='content type', max_length=100)),
            ],
            options={
                'verbose_name_plural': 'attachments',
                'verbose_name': 'attachment',
            },
        ),
        migrations.CreateModel(
            name='EmailMessage',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at')),
                ('sent_at', models.DateTimeField(blank=True, editable=False, null=True, verbose_name='sent at')),
                ('recipient', models.EmailField(max_length=254, verbose_name='recipient')),
                ('sender', models.EmailField(max_length=254, verbose_name='sender')),
                ('sender_name', models.CharField(blank=True, max_length=250, null=True, verbose_name='sender name')),
                ('subject', models.TextField(verbose_name='subject')),
                ('content', models.TextField(verbose_name='content')),
                ('template_slug',
                 models.SlugField(blank=True, editable=False, max_length=100, null=True, verbose_name='slug')),
                ('state',
                 models.IntegerField(choices=[(1, 'waiting'), (2, 'sending'), (3, 'sent'), (4, 'error'), (5, 'debug')],
                                     editable=False, verbose_name='state')),
                ('backend', models.CharField(blank=True, editable=False, null=True, verbose_name='backend',
                                             max_length=250)),
                ('error', models.TextField(blank=True, editable=False, null=True, verbose_name='error')),
                ('extra_data',
                 jsonfield.fields.JSONField(blank=True, editable=False, null=True, verbose_name='extra data')),
                ('extra_sender_data',
                 jsonfield.fields.JSONField(blank=True, editable=False, null=True, verbose_name='extra sender data')),
                ('tag', models.SlugField(blank=True, editable=False, null=True, verbose_name='tag')),
            ],
            options={
                'ordering': ('-created_at',),
                'verbose_name': 'e-mail message',
                'verbose_name_plural': 'e-mail messages'
            }
        ),
        migrations.CreateModel(
            name='EmailRelatedObject',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at')),
                ('object_id', models.TextField(verbose_name='ID of the related object')),
                ('object_id_int', models.PositiveIntegerField(db_index=True, blank=True, null=True,
                                                              verbose_name='ID of the related object in int format')),
                ('content_type',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType',
                                   verbose_name='content type of the related object')),
                ('email_message',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='related_objects',
                                   to='pymess.EmailMessage', verbose_name='e-mail message')),
            ],
            options={
                'ordering': ('-created_at',),
                'verbose_name': 'related object of a e-mail message',
                'verbose_name_plural': 'related objects of e-mail messages',
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='EmailTemplate',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at')),
                ('slug', models.SlugField(editable=False, max_length=100, verbose_name='slug')),
                ('sender', models.EmailField(blank=True, max_length=200, null=True, verbose_name='sender')),
                ('sender_name', models.CharField(blank=True, max_length=250, null=True, verbose_name='sender name')),
                ('body', models.TextField(null=True, verbose_name='message body')),
                ('subject', models.TextField(verbose_name='subject')),
            ],
            options={
                'ordering': ('-created_at',),
                'verbose_name_plural': 'e-mail templates',
                'verbose_name': 'e-mail template',
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='OutputSMSMessage',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at')),
                ('sent_at', models.DateTimeField(blank=True, editable=False, null=True, verbose_name='sent at')),
                ('sender', models.CharField(blank=True, max_length=20, null=True, verbose_name='sender')),
                ('recipient', models.CharField(max_length=20, verbose_name='recipient')),
                ('content', models.TextField(max_length=700, verbose_name='content')),
                ('template_slug',
                 models.SlugField(blank=True, editable=False, max_length=100, null=True, verbose_name='slug')),
                ('state', models.IntegerField(
                    choices=[(1, 'waiting'), (2, 'unknown'), (3, 'sending'), (4, 'sent'), (5, 'error'), (6, 'debug'),
                             (7, 'delivered')], editable=False, verbose_name='state')),
                ('backend', models.CharField(blank=True, editable=False, null=True, verbose_name='backend',
                                             max_length=250)),
                ('error', models.TextField(blank=True, editable=False, null=True, verbose_name='error')),
                ('extra_data',
                 jsonfield.fields.JSONField(blank=True, editable=False, null=True, verbose_name='extra data')),
                ('extra_sender_data',
                 jsonfield.fields.JSONField(blank=True, editable=False, null=True, verbose_name='extra sender data')),
                ('tag', models.SlugField(blank=True, editable=False, null=True, verbose_name='tag')),
            ],
            options={
                'verbose_name_plural': 'output SMS',
                'verbose_name': 'output SMS',
                'ordering': ('-created_at',),
            },
        ),
        migrations.CreateModel(
            name='OutputSMSRelatedObject',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at')),
                ('object_id', models.TextField(verbose_name='ID of the related object')),
                ('object_id_int', models.PositiveIntegerField(db_index=True, blank=True, null=True,
                                                              verbose_name='ID of the related object in int format')),
                ('content_type',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType',
                                   verbose_name='content type of the related object')),
                ('output_sms_message', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE,
                                                         related_name='related_objects', to='pymess.OutputSMSMessage',
                                                         verbose_name='output SMS message')),
            ],
            options={
                'ordering': ('-created_at',),
                'verbose_name': 'related object of a SMS message',
                'verbose_name_plural': 'related objects of SMS messages',
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='SMSTemplate',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at')),
                ('slug', models.SlugField(editable=False, max_length=100, verbose_name='slug')),
                ('body', models.TextField(null=True, verbose_name='message body')),
            ],
            options={
                'ordering': ('-created_at',),
                'verbose_name_plural': 'SMS templates',
                'verbose_name': 'SMS template',
                'abstract': False,
            },
        ),
        migrations.RemoveField(
            model_name='userdevice',
            name='user',
        ),
        migrations.DeleteModel(
            name='UserDevice',
        ),
        migrations.AddField(
            model_name='outputsmsmessage',
            name='template',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL,
                                    related_name='output_sms_messages', to=settings.SMS_TEMPLATE_MODEL,
                                    verbose_name='template'),
        ),
        migrations.AddField(
            model_name='emailmessage',
            name='template',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL,
                                    related_name='email_messages', to=settings.EMAIL_TEMPLATE_MODEL,
                                    verbose_name='template'),
        ),
        migrations.AddField(
            model_name='attachment',
            name='email_message',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='attachments',
                                    to='pymess.EmailMessage', verbose_name='e-mail message'),
        ),
    ]



import logging

from datetime import timedelta

from django.utils import timezone
from django.utils.encoding import force_text
from django.utils.translation import ugettext

from chamber.exceptions import PersistenceException

from pymess.config import settings, get_sms_template_model, get_sms_sender
from pymess.models import OutputSMSMessage
from pymess.utils import fullname


LOGGER = logging.getLogger(__name__)


class SMSBackend(object):
    """
    Base class for SMS backend containing implementation of concrete SMS service that
    is used for sending SMS messages.
    """

    class SMSSendingError(Exception):
        pass

    def _get_extra_sender_data(self):
        """
        Gets arguments that will be saved with the sms message in the extra_sender_data field
        """
        return {}

    def _get_extra_message_kwargs(self):
        """
        Gets model message kwargs that will be saved with the sms message
        """
        return {}

    def update_message(self, message, extra_sender_data=None, **kwargs):
        """
        Method for updating state of the message after sending
        :param message: SMS message object
        :param extra_sender_data: extra data that will be saved to the extra_sender_data field
        :param kwargs: changed object kwargs
        :return:
        """
        extra_sender_data = {
            **self._get_extra_sender_data(),
            **({} if extra_sender_data is None else extra_sender_data)
        }
        message.change_and_save(
            backend=fullname(self),
            extra_sender_data=extra_sender_data,
            **kwargs
        )

    def create_message(self, recipient, content, related_objects, tag, template, **sms_kwargs):
        """
        Create SMS which will be logged in the database.
        :param recipient: phone number of the recipient
        :param content: content of the SMS message
        :param related_objects: list of related objects that will be linked with the SMS message using generic
        relation
        :param tag: string mark that will be saved with the message
        :param template: template object from which content of the message was created
        :param sms_kwargs: extra attributes that will be saved with the message
        """
        try:
            message = OutputSMSMessage.objects.create(
                recipient=recipient,
                content=content,
                state=self.get_initial_sms_state(recipient),
                extra_data=sms_kwargs,
                tag=tag,
                template=template,
                template_slug=template.slug if template else None,
                **self._get_extra_message_kwargs()
            )
            if related_objects:
                message.related_objects.create_from_related_objects(*related_objects)
            return message
        except PersistenceException as ex:
            raise self.SMSSendingError(force_text(ex))

    def publish_message(self, message):
        """
        Place for implementation logic of sending SMS message.
        :param message: SMS message instance
        """
        raise NotImplementedError

    def publish_messages(self, messages):
        """
        Sends more SMS messages together. If concrete SMS backend provides send more messages at once the method
        can be overridden
        :param messages: list of SMS messages
        """
        [self.publish_message(message) for message in messages]

    def send(self, recipient, content, related_objects=None, tag=None, template=None, **sms_kwargs):
        """
        Send SMS with the text content to the phone number (recipient)
        :param recipient: phone number of the recipient
        :param content: text content of the message
        :param related_objects: list of related objects that will be linked with the SMS message using generic
        relation
        :param tag: string mark that will be saved with the message
        :param template: template object from which content of the message was create
        :param sms_kwargs: extra attributes that will be stored to the message
        """
        message = self.create_message(recipient, content, related_objects, tag, template, **sms_kwargs)
        self.publish_message(message)
        return message

    def bulk_send(self, recipients, content, related_objects=None, tag=None, template=None, **sms_kwargs):
        """
        Send more SMS messages in one bulk
        :param recipients: list of phone numbers of recipients
        :param content: text content of the messages
        :param related_objects: list of related objects that will be linked with the SMS message using generic
        relation
        :param tag: string mark that will be saved with the message
        :param template: template object from which content of the message was create
        :param sms_kwargs: extra attributes that will be stored with messages
        """
        messages = [
            self.create_message(recipient, content, related_objects, tag, template, **sms_kwargs)
            for recipient in recipients
        ]
        self.publish_messages(messages)
        return messages

    def get_initial_sms_state(self, recipient):
        """
        returns initial state for logged SMS instance.
        :param recipient: phone number of the recipient
        """
        return OutputSMSMessage.STATE.WAITING

    def _update_sms_states(self, messages):
        """
        If SMS sender provides check SMS delivery this method can be overridden.
        :param messages: messages which state will be updated
        """
        raise NotImplementedError('Check SMS state is not supported with the backend')

    def bulk_check_sms_states(self):
        """
        Method that find messages that is not in the final state and updates its states.
        """
        messages_to_check = OutputSMSMessage.objects.filter(state=OutputSMSMessage.STATE.SENDING)
        if messages_to_check.exists():
            self._update_sms_states(messages_to_check)

        idle_output_sms = OutputSMSMessage.objects.filter(
            state=OutputSMSMessage.STATE.SENDING,
            created_at__lt=timezone.now() - timedelta(minutes=settings.SMS_IDLE_MESSAGES_TIMEOUT_MINUTES)
        )
        if settings.SMS_LOG_IDLE_MESSAGES and idle_output_sms.exists():
            LOGGER.warning('{count_sms} Output SMS is more than {timeout} minutes in state "SENDING"'.format(
                count_sms=idle_output_sms.count(), timeout=settings.SMS_IDLE_MESSAGES_TIMEOUT_MINUTES
            ))

        if settings.SMS_SET_ERROR_TO_IDLE_MESSAGES:
            idle_output_sms.update(
                state=OutputSMSMessage.STATE.ERROR, error=ugettext('timeouted')
            )


def send_template(recipient, slug, context_data, related_objects=None, tag=None):
    """
    Helper for building and sending SMS message from a template.
    :param recipient: phone number of the recipient
    :param slug: slug of a SMS template
    :param context_data: dict of data that will be sent to the template renderer
    :param related_objects: list of related objects that will be linked with the SMS message using generic
        relation
    :param tag: string mark that will be saved with the message
    :return: SMS message object or None if template cannot be sent
    """
    return get_sms_template_model().objects.get(slug=slug).send(
        recipient,
        context_data,
        related_objects=related_objects,
        tag=tag
    )


def send(recipient, content, related_objects=None, tag=None, **sms_kwargs):
    """
    Helper for sending SMS message.
    :param recipient: phone number of the recipient
    :param content: text content of the messages
    :param related_objects:
    :param tag: string mark that will be saved with the message
    :param sms_kwargs: extra attributes that will be stored with messages
    :return: True if SMS was successfully sent or False if message is in error state
    """
    return get_sms_sender().send(
        recipient,
        content,
        related_objects=related_objects,
        tag=tag,
        **sms_kwargs
    ).failed